unit Debug;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, Mask;

type
  TDebugForm = class(TForm)

    ListBox: TShape;
    ListScrollBar: TScrollBar;
    StaticTextList: TStaticText;

    RegBox: TShape;
    RegScrollBar: TScrollBar;
    StaticTextReg: TStaticText;

    HexBox: TShape;
    HexScrollBar: TScrollBar;
    StaticTextHex: TStaticText;

    StepGroup: TGroupBox;
    StepButton: TButton;

    TraceGroup: TGroupBox;
    TraceEdit: TMaskEdit;
    TraceButton: TButton;

    BpGroup: TGroupBox;
    BpEdit: TMaskEdit;
    BpButton: TButton;

{ DISASSEMBLY BOX EVENTS }
    procedure ListScrollBarChange(Sender: TObject);
    procedure ListBoxMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ListBoxKeyPress(Sender: TObject; var Key: Char);

{ REGISTERS BOX EVENTS }
    procedure RegBoxMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure RegBoxKeyPress(Sender: TObject; var Key: Char);

{ HEX EDITOR BOX EVENTS }
    procedure HexScrollBarChange(Sender: TObject);
    procedure HexBoxMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure HexBoxKeyPress(Sender: TObject; var Key: Char);

{ GENERAL FORM EVENTS }
    procedure DebugCreate(Sender: TObject);
    procedure DebugShow(Sender: TObject);
    procedure DebugPaint(Sender: TObject);
    procedure DebugHide(Sender: TObject);

{ MACHINE CODE EXECUTION CONTROL EVENTS }
    procedure StepButtonClick(Sender: TObject);
    procedure TraceEditChange(Sender: TObject);
    procedure TraceButtonClick(Sender: TObject);
    procedure BpEditChange(Sender: TObject);
    procedure BpButtonClick(Sender: TObject);
    procedure CallCloseEdit(Sender: TObject);

  private
    { Private declarations }
  public
    { Public declarations }
  end;


var
  DebugForm: TDebugForm;


implementation

uses Make_def, Pdp11dis, Cpu;

{$R *.DFM}

const
  BOXPADDING = 10;	{space between the border of the box and the contents}

var
  HexAddr: word;
  ListAddr: word;
  OldListScrollBarPosition: integer;

  EditState: (NoEdit, ListAddrEdit, RegEdit, PSWEdit, HexAddrEdit, HexDataEdit);
  edstr: string[4]; {string edited by BoxEdit}
  EditAddr: word;	{address of the edited object - memory location, register}


{ GENERAL DISPLAY ROUTINES }

{ clears a box }
procedure BoxClear (Sh: TShape);
var
  BoxRect: TRect;
  P: integer;
begin
  with Sh do
  begin
    P := Pen.Width+BOXPADDING;
    BoxRect := Rect(Left+P, Top+P, Left+Width-P, Top+Height-P);
  end {with};
  with DebugForm do
  begin
    Canvas.Brush.Style := bsSolid;
    Canvas.Brush.Color := Color;
    Canvas.FillRect (BoxRect);
  end {with};
end {BoxClear};


{ displays a string in a box }
procedure BoxText (Sh: TShape; Col, Row: integer; s: string);
var
  P, W, H: integer;
  BoxRect: TRect;
begin
  with Sh do
  begin
    P := Pen.Width+BOXPADDING;
    BoxRect := Rect(Left+P, Top+P, Left+Width-P, Top+Height-P);
  end {with};
  with DebugForm.Canvas do
  begin
    H := Abs(Font.Height);
    W := TextWidth('0');
    Brush.Style := bsClear;
    TextRect (BoxRect, BoxRect.Left+Col*W, BoxRect.Top+Row*H, s);
  end {with};
end {BoxText};


{ returns count of rows }
function BoxRowsCount (Sh: TShape) : integer;
begin
  BoxRowsCount := (Sh.Height - 2*(Sh.Pen.Width+BOXPADDING))
    div Abs(DebugForm.Canvas.Font.Height);
end;


{ expects mouse pointer coordinate X, returns column }
function BoxCol (Sh: TShape; X: integer) : integer;
var
  W: integer;
begin
  W := DebugForm.Canvas.Pen.Width+BOXPADDING;
  if (X < W) or (X > Sh.Width-W) then BoxCol := 0
  else BoxCol := (X - W) div DebugForm.Canvas.TextWidth('0');
end;


{ expects mouse pointer coordinate Y, returns row }
function BoxRow (Sh: TShape; Y: integer) : integer;
var
  W: integer;
begin
  W := DebugForm.Canvas.Pen.Width+BOXPADDING;
  if (Y < W) or (Y > Sh.Height-W) then BoxRow := 0
  else BoxRow := (Y - W) div Abs(DebugForm.Canvas.Font.Height);
end;


procedure BoxCursor (Sh: TShape; Col, Row, Size: integer; cl: TColor);
var
  P, W, H, X, Y: integer;
begin
  with DebugForm.Canvas do
  begin
    H := Abs(Font.Height);
    W := TextWidth('0');
    Brush.Style := bsSolid;
    Brush.Color := cl;
  end {with};
  with Sh do
  begin
    P := Pen.Width+BOXPADDING;
    X := Left + P + Col*W;
    Y := Top + P + Row*H;
  end {with};
  DebugForm.Canvas.FillRect (Rect(X, Y, X+W*Size, Y+H));
end {BoxCursor};


procedure BoxEdit (Sh: TShape; Col, Row: integer; Entry: char);
var
  i, Size: integer;
begin
  Size := Length(edstr);
  if Ord(Entry) = 8 then {backspace}
  begin
    for i := Size downto 2 do edstr[i] := edstr[i-1];
    edstr[1] := ' ';
  end
  else
  begin
    for i := 2 to Size do edstr[i-1] := edstr[i];
    edstr[Size] := Entry;
  end {if};
  BoxCursor (Sh, Col, Row, Size, clYellow);
  BoxText (Sh, Col, Row, Copy(edstr, 1, Size));
end {BoxEdit};



{ DISASSEMBLY BOX ROUTINES }

procedure DisplayList;
var
  i, rows, index: integer;
  opcode: word;
  c: char;
  ListStartAddr, ListEndAddr: word;
begin
  ListAddr := ListAddr and $FFFE;
  if (ListAddr < ROMEND) then
  begin
    ListStartAddr := ROMSTART;
    ListEndAddr := ROMEND;
    DebugForm.ListScrollBar.Max := (ROMEND-ROMSTART) div 2;
  end
  else if (ListAddr >= RAMSTART) and (ListAddr < RamEnd) then
  begin
    ListStartAddr := RAMSTART;
    ListEndAddr := RamEnd;
    DebugForm.ListScrollBar.Max := (RamEnd-RAMSTART) div 2;
  end
  else {out of allowed address space}
  begin
    ListAddr := ROMSTART;
    ListStartAddr := ROMSTART;
    ListEndAddr := ROMEND;
  end {if};
  loc_addr := ListAddr;
  with DebugForm do
  begin
    BoxClear (ListBox);
    rows := BoxRowsCount (ListBox);
    for i := 0 to rows-1 do
    begin
      if loc_addr >= ListEndAddr then break;
      if (EditState = ListAddrEdit) and (i = 0) then
      begin
        BoxCursor (ListBox, 0, 0, 4, clYellow);
        BoxText (ListBox, 0, 0, edstr);
      end
      else BoxText (ListBox, 0, i, IntToHex(loc_addr, 4));
      BoxText (ListBox, 4, i, ':');
      opcode := FetchWord;
      index := ScanNTAB (opcode);
      if (opcode and $8000) = 0 then c := ' ' else c := 'b';
      BoxText (ListBox, 6, i, Mnemonic (index, c));
      BoxText (ListBox, 12, i, Arguments (index, opcode));
    end {for};
    OldListScrollBarPosition := (ListAddr-ListStartAddr) div 2;
    ListScrollBar.Position := OldListScrollBarPosition;
    if loc_addr < ListEndAddr then
    begin
      ListScrollBar.LargeChange := (loc_addr-ListAddr) div 2;
      ListScrollBar.PageSize := (loc_addr-ListAddr) div 2;
    end;
  end {with};
end {DisplayList};



{ REGISTERS BOX ROUTINES }

const
  regname: array[0..8] of string[4] =
    ('R0:', 'R1:', 'R2:', 'R3:', 'R4:', 'R5:', 'SP:', 'PC:', 'PSW:');
  flagname: array[0..5] of char =
    ('I', 'T', 'N', 'Z', 'V', 'C');
  flagmask: array[0..5] of word =
    ($0100, $0010, $0008, $0004, $0002, $0001);

procedure DisplayRegisters;
var
  i: integer;
  c: char;
begin
  with DebugForm do
  begin
    BoxClear (RegBox);
    for i := 0 to 8 do
    begin
      BoxText (RegBox, 0, i, regname[i]);
      if (EditState = RegEdit) and (i = EditAddr) then
      begin
        BoxCursor (RegBox, 4, i, 4, clYellow);
        BoxText (RegBox, 4, i, edstr);
      end
      else if i <> 8 then
        BoxText (RegBox, 4, i, IntToHex(ptrw(@reg[i*2])^, 4))
      else
        BoxText (RegBox, 4, i, IntToHex(psw, 4));
    end {for};
    for i := 0 to 5 do
    begin
      if (psw and flagmask[i]) = 0 then c := '-' else c := flagname[i];
      BoxText (RegBox, i, 9, c);
    end {for};
  end {with};
end;



{ HEX EDITOR BOX ROUTINES }

function HexLine (a: word) : string;
var
  i: word;
begin
  Result := '';
  for i := 0 to 15 do
    Result := Result + IntToHex(ram[a+i-RAMSTART],2) + ' ';
end {HexLine};


function CharLine (a: word) : string;
var
  i: word;
  x: byte;
  c: char;
begin
  Result := '';
  for i := 0 to 15 do
  begin
    x := ram[a+i-RAMSTART] and $7F;
    if (x < $20) or (x > $7E) then c := '.' else c := Chr(x);
    Result := Result + c;
  end {for};
end {CharLine};


procedure DisplayHex;
var
  i, rows: integer;
  a: word;
begin
  a := HexAddr;
  with DebugForm do
  begin
    BoxClear (HexBox);
    rows := BoxRowsCount (HexBox);
    for i := 0 to rows-1 do
    begin
      if (EditState = HexAddrEdit) and (i = 0) then
      begin
        BoxCursor (HexBox, 0, 0, 4, clYellow);
        BoxText (HexBox, 0, 0, edstr);
      end
      else BoxText (HexBox, 0, i, IntToHex(a, 4));
      BoxText (HexBox, 4, i, ':');
      BoxText (HexBox, 6, i, HexLine(a));
      BoxText (HexBox, 54, i, CharLine(a));
      if (EditState = HexDataEdit) and (a = (EditAddr and $FFF0)) then
      begin
        BoxCursor (HexBox, (EditAddr and $F)*3 + 6, i, 2, clYellow);
        BoxText (HexBox, (EditAddr and $F)*3 + 6, i, Copy(edstr, 1, 2));
      end {if};
      a := a+16;
      if a > RamEnd then a := RAMSTART;
    end {for};
  end {with};
end {DisplayHex};



procedure CloseEdit;
begin
  EditState := NoEdit;
  DisplayRegisters;
  DisplayHex;
  DisplayList;
end;



{ BOX EVENTS }

procedure TDebugForm.ListScrollBarChange(Sender: TObject);
begin
  if ListScrollBar.Position <> OldListScrollBarPosition then
  begin
    if ListAddr >= RAMSTART then
      ListAddr := 2*ListScrollBar.Position + RAMSTART
    else
      ListAddr := 2*ListScrollBar.Position + ROMSTART;
    DisplayList;
  end {if};
end;


procedure TDebugForm.ListBoxMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Col, Row: integer;
begin
  ListScrollBar.SetFocus;
  CloseEdit;
  Col := BoxCol (HexBox, X);
  Row := BoxRow (HexBox, Y);
  if (Col < 4) and (Row = 0) then
  begin
    EditState := ListAddrEdit;
    EditAddr := ListAddr;
    edstr := '    ';
    DisplayList;
  end {if};
end;


procedure TDebugForm.ListBoxKeyPress(Sender: TObject; var Key: Char);
begin
  if EditState = ListAddrEdit then
  begin
    if GetDigit(Key) < 16 then BoxEdit (ListBox, 0, 0, Key)
    else if Ord(Key) = $1B then CloseEdit
    else if Ord(Key) = $0D then
    begin
      ListAddr := GetValue (edstr, 16);
      CloseEdit;
    end {if};
  end {if};
end;


procedure TDebugForm.RegBoxMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  Col, Row: integer;
begin
  RegScrollBar.SetFocus;
  CloseEdit;
  Col := BoxCol (RegBox, X);
  Row := BoxRow (RegBox, Y);
  if (Col >= 4) and (Col < 8) and (Row < 9) then
  begin
    EditState := RegEdit;
    EditAddr := Row;
    edstr := '    ';
    DisplayRegisters;
  end {if};
end;


procedure TDebugForm.RegBoxKeyPress(Sender: TObject; var Key: Char);
var
  x: word;
begin
  if EditState = RegEdit then
  begin
    if GetDigit(Key) < 16 then BoxEdit (RegBox, 4, EditAddr, Key)
    else if Ord(Key) = $1B then CloseEdit
    else if Ord(Key) = $0D then
    begin
      x := GetValue (edstr, 16);
      if EditAddr < 6 then ptrw(@reg[2*EditAddr])^ := x
      else if EditAddr < 8 then ptrw(@reg[2*EditAddr])^ := x and $FFFE
      else psw := x;
      CloseEdit;
    end {if};
  end {if};
end;


procedure TDebugForm.HexScrollBarChange(Sender: TObject);
begin
  HexAddr := HexScrollBar.Position*16 + RAMSTART;
  DisplayHex;
end;


procedure TDebugForm.HexBoxMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  Col, Row: integer;
begin
  HexScrollBar.SetFocus;
  CloseEdit;
  Col := BoxCol (HexBox, X);
  Row := BoxRow (HexBox, Y);
  if (Row = 0) and (Col < 4) then
  begin {select HexAddr edition}
    EditState := HexAddrEdit;
    EditAddr := 0;
    edstr := '    ';
    DisplayHex;
  end
  else if (Col >= 6) and (Col < 54) and ((Col mod 3) < 2) then
  begin {select byte edition in the HexBox}
    EditState := HexDataEdit;
    EditAddr := HexAddr + 16*Row + Col div 3 - 2;
    edstr := '  ';
    DisplayHex;
  end {if};
end;


procedure TDebugForm.HexBoxKeyPress(Sender: TObject; var Key: Char);
var
  x: word;
  Col, Row: integer;
begin
  if EditState = HexAddrEdit then
  begin
    if GetDigit(Key) < 16 then BoxEdit (HexBox, 0, 0, Key)
    else if Ord(Key) = $1B then CloseEdit
    else if Ord(Key) = $0D then
    begin
      x := GetValue (edstr, 16) and $FFF0;
      if (x >= RAMSTART) and (x <= RamEnd - 16*BoxRowsCount(HexBox)) then
        HexAddr := x;
      CloseEdit;
    end {if};
  end
  else if EditState = HexDataEdit then
  begin
    if GetDigit(Key) < 16 then
    begin
      Col := ((EditAddr - HexAddr) mod 16) * 3 + 6;
      Row := (EditAddr - HexAddr) div 16;
      BoxEdit (HexBox, Col, Row, Key);
    end
    else if Ord(Key) = $1B then CloseEdit
    else if Ord(Key) = $0D then
    begin
      ram[EditAddr-RAMSTART] := GetValue (edstr, 16);
      CloseEdit;
    end {if};
  end {if};
end;



{ GENERAL FORM EVENTS }

procedure TDebugForm.DebugCreate(Sender: TObject);
begin
  EditState := NoEdit;
  HexAddr := RAMSTART;
  ListAddr := ROMSTART;
  OldListScrollBarPosition := 0;
  HexScrollBar.Max := (RamEnd-RAMSTART) div 16 - BoxRowsCount(HexBox);
  HexScrollBar.LargeChange := BoxRowsCount(HexBox);
  HexScrollBar.PageSize := BoxRowsCount(HexBox);
  ListScrollBar.Max := (ROMEND-ROMSTART) div 2;
end;


procedure TDebugForm.DebugShow(Sender: TObject);
begin
  CpuStop := True;
  CpuSteps := -1;
  BreakPoint := -1;
  ListAddr := ptrw(@reg[R7])^;
end;


procedure TDebugForm.DebugPaint(Sender: TObject);
begin
  DisplayRegisters;
  DisplayHex;
  DisplayList;
end;


procedure TDebugForm.DebugHide(Sender: TObject);
begin
  Hide;
  CpuStop := False;
end;



{ MACHINE CODE EXECUTION CONTROL }

procedure TDebugForm.StepButtonClick(Sender: TObject);
begin
  EditState := NoEdit;
  CpuRun;
  DisplayRegisters;
  DisplayHex;
  ListAddr := ptrw(@reg[R7])^;
  DisplayList;
end;


procedure TDebugForm.TraceEditChange(Sender: TObject);
var
  i: integer;
  c: char;
  s: string[4];
begin
  with DebugForm.TraceEdit do
  begin
    s := '';
    i := 1;
    while i <= Length(Text) do
    begin
      c := Text[i];
      if GetDigit(c) <= 9 then s := s+c;
      i := i+1;
    end {while};
    Text := s;
  end {with};
end;


procedure TDebugForm.TraceButtonClick(Sender: TObject);
var
  i: integer;
begin
  EditState := NoEdit;
  i := GetValue (DebugForm.TraceEdit.Text, 10);
  if i > 0 then
  begin
    BreakPoint := -1;
    CpuSteps := i;
    Hide;
  end {if};
end;


procedure TDebugForm.BpEditChange(Sender: TObject);
var
  i: integer;
  c: char;
  s: string[4];
begin
  with DebugForm.BPEdit do
  begin
    s := '';
    i := 1;
    while i <= Length(Text) do
    begin
      c := Text[i];
      if GetDigit(c) <= 15 then s := s+c;
      i := i+1;
    end {while};
    Text := s;
  end {with};
end;


procedure TDebugForm.BpButtonClick(Sender: TObject);
var
  i: integer;
begin
  EditState := NoEdit;
  i := GetValue (DebugForm.BpEdit.Text, 16);
  BreakPoint := i;
  CpuSteps := -1;
  Hide;
end;


procedure TDebugForm.CallCloseEdit(Sender: TObject);
begin
  CloseEdit;
end;

end.
